from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, Protocol, Self, TypeVar, runtime_checkable

from waku.di._context import InjectionContext
from waku.di._inject import context_var
from waku.di._utils import guess_return_type

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from contextlib import AbstractAsyncContextManager, AbstractContextManager
    from types import TracebackType

    from waku.di._types import FactoryType

__all__ = [
    'DependencyProvider',
    'InjectionContext',
    'Object',
    'Provider',
    'Scoped',
    'Singleton',
    'Transient',
]


_T = TypeVar('_T')


@runtime_checkable
class Provider(Hashable, Protocol[_T]):
    impl: Any
    type_: type[_T]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}[{self.type_!r}]'


class Scoped(Provider[_T]):
    def __init__(
        self,
        factory: FactoryType[_T],
        type_: type[_T] | None = None,
    ) -> None:
        self.impl = factory
        self.type_ = type_ or guess_return_type(factory)

    def __hash__(self) -> int:
        return hash(self.type_)


class Singleton(Scoped[_T]):
    pass


class Transient(Scoped[_T]):
    pass


class Object(Provider[_T]):
    impl: _T

    def __init__(
        self,
        object_: _T,
        type_: type[_T] | None = None,
    ) -> None:
        self.type_ = type_ or type(object_)
        self.impl = object_

    def __hash__(self) -> int:
        return hash(self.type_)


class DependencyProvider(ABC):
    _exit_stack: contextlib.AsyncExitStack

    @abstractmethod
    def register(self, *providers: Provider[Any]) -> None: ...

    @contextlib.asynccontextmanager
    async def context(self) -> AsyncIterator[InjectionContext]:
        if ctx := context_var.get(None):
            yield ctx
        else:
            async with self._context() as ctx:
                token = context_var.set(ctx)
                try:
                    yield ctx
                finally:
                    context_var.reset(token)

    @abstractmethod
    def override(self, provider: Provider[Any]) -> AbstractContextManager[None]: ...

    @abstractmethod
    def _lifespan(self) -> AbstractAsyncContextManager[None]: ...

    @abstractmethod
    def _context(self) -> InjectionContext: ...

    async def __aenter__(self) -> Self:
        await self._exit_stack.__aenter__()
        await self._exit_stack.enter_async_context(self._lifespan())
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)



from __future__ import annotations

from typing import TYPE_CHECKING

from waku.ext.mediator.handlers.dispatcher import RequestDispatcher

if TYPE_CHECKING:
    from waku.di import DependencyProvider
    from waku.ext.mediator.handlers.handler import Request, ResponseT
    from waku.ext.mediator.handlers.map import RequestMap
    from waku.ext.mediator.middlewares import MiddlewareChain


__all__ = ['Mediator']


class Mediator:
    def __init__(
        self,
        *,
        request_map: RequestMap,
        dependency_provider: DependencyProvider,
        middleware_chain: MiddlewareChain | None = None,
        dispatcher_class: type[RequestDispatcher] | None = None,
    ) -> None:
        dispatcher_class = dispatcher_class or RequestDispatcher
        self._dispatcher = dispatcher_class(
            request_map=request_map,
            dependency_provider=dependency_provider,
            middleware_chain=middleware_chain,
        )

    async def send(self, request: Request[ResponseT]) -> ResponseT:
        dispatch_result = await self._dispatcher.dispatch(request)
        # TODO: dispatch events from result  # noqa: FIX002
        return dispatch_result.response



from __future__ import annotations

import dataclasses
import warnings
from typing import TYPE_CHECKING, Final

from waku.extensions import OnApplicationInit

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.application import Application
    from waku.ext.validation import ValidationRule
    from waku.ext.validation._errors import ValidationError


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class ValidationContext:
    app: Application


class ValidationExtension(OnApplicationInit):
    def __init__(self, rules: Sequence[ValidationRule], *, strict: bool = True) -> None:
        self.rules = rules
        self.strict: Final = strict

    def on_app_init(self, app: Application) -> None:
        context = ValidationContext(app=app)

        for rule in self.rules:
            if err := rule.validate(context):
                self._raise(err=err)

    def _raise(self, err: ValidationError) -> None:
        if self.strict:
            raise err
        warnings.warn(str(err), stacklevel=3)



from __future__ import annotations

import re
from typing import NewType

import pytest

from tests.mock import DummyDI
from waku.application import Application
from waku.di import Scoped
from waku.ext.validation import ValidationError, ValidationExtension, ValidationRule
from waku.module import Module


class A:
    pass


class B:
    def __init__(self, a: A) -> None:
        self._a = a


C = NewType('C', A)


class D:
    def __init__(self, c: C) -> None:
        self._c = c


def _impl() -> int:
    return 1


@pytest.mark.parametrize(
    ('imports', 'exports'),
    [
        (False, False),
        (False, True),
        (True, False),
    ],
)
def test_inaccessible(
    imports: bool,
    exports: bool,
    rules: list[ValidationRule],
) -> None:
    a = Module(name='A', providers=[Scoped(A)], exports=[A] if exports else [])
    b = Module(name='B', providers=[Scoped(B)], imports=[a] if imports else [])

    error_message = f"{b!r} depends on {A!r} but it's not accessible to it"

    with pytest.raises(ValidationError) as exc_info:
        Application(
            'app',
            modules=[a, b],
            dependency_provider=DummyDI(),
            extensions=[ValidationExtension(rules)],
        )
    assert str(exc_info.value) == error_message

    with pytest.warns(Warning, match=re.escape(error_message)):
        Application(
            name='app',
            modules=[a, b],
            dependency_provider=DummyDI(),
            extensions=[ValidationExtension(rules, strict=False)],
        )


def test_ok(rules: list[ValidationRule]) -> None:
    a = Module(name='A', providers=[Scoped(A), Scoped(_impl, C)], exports=[A, C])
    b = Module(name='B', providers=[Scoped(B), Scoped(D)], imports=[a])
    Application(
        'app',
        modules=[a, b],
        dependency_provider=DummyDI(),
        extensions=[ValidationExtension(rules)],
    )


def test_ok_with_global_providers(rules: list[ValidationRule]) -> None:
    a = Module(name='A', providers=[Scoped(A)], is_global=True)
    b = Module(name='B', providers=[Scoped(B)])
    Application(
        'app',
        modules=[a, b],
        dependency_provider=DummyDI(),
        extensions=[ValidationExtension(rules)],
    )


def test_ok_with_application_providers(rules: list[ValidationRule]) -> None:
    b = Module(name='B', providers=[Scoped(B)], exports=[B])
    Application(
        'app',
        modules=[b],
        dependency_provider=DummyDI(),
        extensions=[ValidationExtension(rules)],
        providers=[Scoped(A)],
    )



import dataclasses
from typing import Any

import pytest

from tests.mock import DummyDI
from waku import Application, Module
from waku.di import Object, Provider, Scoped, Singleton, Transient
from waku.ext.validation import ValidationError, ValidationExtension, ValidationRule


class _A:
    pass


@dataclasses.dataclass
class _B:
    a: _A


@pytest.mark.parametrize(
    ('provider', 'dependency', 'should_error'),
    [
        (Scoped(_B), Scoped(_A), False),
        (Scoped(_B), Singleton(_A), False),
        (Singleton(_B), Singleton(_A), False),
        (Singleton(_B), Scoped(_A), True),
        (Singleton(_B), Transient(_A), True),
        # "Object" should already be instantiated
        (Object(_B(a=_A())), Scoped(_A), False),
        (Object(_B(a=_A())), Transient(_A), False),
        (Object(_B(a=_A())), Singleton(_A), False),
        (Scoped(_B), Object(_A()), False),
        (Transient(_B), Object(_A()), False),
        (Singleton(_B), Object(_A()), False),
    ],
)
def test_scope_mismatch(
    provider: Provider[Any],
    dependency: Provider[Any],
    should_error: bool,
    rules: list[ValidationRule],
) -> None:
    def create_app() -> None:
        Application(
            'app',
            modules=[Module(name='module', providers=[provider, dependency])],
            dependency_provider=DummyDI(),
            extensions=[ValidationExtension(rules)],
        )

    if not should_error:
        create_app()
        return

    with pytest.raises(ValidationError) as exc_info:
        create_app()

    assert str(exc_info.value).startswith(f'{provider!r} depends on {dependency!r}')



from waku.ext.validation._abc import ValidationRule
from waku.ext.validation._errors import ValidationError
from waku.ext.validation._extension import ValidationExtension

__all__ = ['ValidationError', 'ValidationExtension', 'ValidationRule']



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ParamSpec, Protocol, Self, TypeVar, overload

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Iterable
    from types import TracebackType

    from waku.di._utils import Dependency

_T = TypeVar('_T')
_P = ParamSpec('_P')


class InjectionContext(Protocol):
    async def resolve(self, type_: type[_T]) -> _T: ...

    @overload
    async def execute(
        self,
        function: Callable[_P, Coroutine[Any, Any, _T]],
        dependencies: Iterable[Dependency[object]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T: ...

    @overload
    async def execute(
        self,
        function: Callable[_P, _T],
        dependencies: Iterable[Dependency[object]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T: ...

    async def execute(
        self,
        function: Callable[_P, Coroutine[Any, Any, _T] | _T],
        dependencies: Iterable[Dependency[object]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T: ...

    async def __aenter__(self) -> Self: ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None: ...



from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine, Iterator
from typing import Any, TypeAlias, TypeVar

__all__ = ['FactoryType']

_T = TypeVar('_T')


FactoryType: TypeAlias = (
    type[_T]
    | Callable[..., _T]
    | Callable[..., Awaitable[_T]]
    | Callable[..., Coroutine[Any, Any, _T]]
    | Callable[..., Iterator[_T]]
    | Callable[..., AsyncIterator[_T]]
)



from __future__ import annotations

import functools
import inspect
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, overload

from waku.di._utils import clear_wrapper, collect_dependencies

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, AsyncIterator, Callable, Coroutine

    from waku.di._context import InjectionContext


__all__ = [
    'context_var',
    'inject',
]

_T = TypeVar('_T')
_P = ParamSpec('_P')

context_var: ContextVar[InjectionContext] = ContextVar('waku_context')


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    wrapper = _inject(function)
    return clear_wrapper(wrapper)


def _wrap_async(
    function: Callable[_P, Coroutine[Any, Any, _T]],
) -> Callable[_P, Coroutine[Any, Any, _T]]:
    dependencies = list(collect_dependencies(function))

    @functools.wraps(function)
    async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        context = context_var.get()
        execute = context.execute(
            function,
            dependencies,
            *args,
            **kwargs,
        )

        return await execute

    return wrapper


def _wrap_async_gen(
    function: Callable[_P, Coroutine[Any, Any, AsyncIterable[_T]]],
) -> Callable[_P, AsyncIterable[_T]]:
    wrapped = _wrap_async(function)

    @functools.wraps(function)
    async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> AsyncIterator[_T]:
        async for element in await wrapped(*args, **kwargs):
            yield element

    return wrapper


@overload
def _inject(func: Callable[_P, _T]) -> Callable[_P, _T]: ...


@overload
def _inject() -> Callable[[Callable[_P, _T]], Callable[_P, _T]]: ...


def _inject(
    func: Callable[_P, _T] | None = None,
) -> Callable[_P, _T] | Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    def wrap(function: Callable[_P, _T]) -> Callable[_P, _T]:
        if inspect.iscoroutinefunction(function):
            return _wrap_async(function)  # type: ignore[return-value]

        if inspect.isasyncgenfunction(function):
            return _wrap_async_gen(  # type: ignore[return-value]
                function,  # type: ignore[arg-type]
            )

        msg = 'inject decorator cannot be used with sync functions'
        raise NotImplementedError(msg)

    if func is None:
        return wrap

    return wrap(func)



from dataclasses import dataclass, field
from datetime import datetime
from typing import Generic, TypeVar

_TPrimaryKey = TypeVar('_TPrimaryKey')

_sentinel = object()


@dataclass(kw_only=True)
class Entity(Generic[_TPrimaryKey]):  # noqa: PLW1641
    id: _TPrimaryKey = field(default=_sentinel, hash=True)  # type: ignore[assignment]

    @property
    def is_transient(self) -> bool:
        return self.id is _sentinel

    def __eq__(self, other: object) -> bool:
        if self is not other and self.is_transient:
            return False
        return super().__eq__(other)


@dataclass(kw_only=True, frozen=True)
class EventData:
    event_time: datetime


@dataclass(kw_only=True)
class AggregateRoot(Entity[_TPrimaryKey]):
    domain_events: list[EventData] = field(default_factory=list)



from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager, contextmanager, nullcontext
from typing import TYPE_CHECKING, Any, cast

import aioinject
from aioinject.context import context_var as aioinject_context

from waku.di import DependencyProvider, InjectionContext, Object, Provider, Scoped, Singleton, Transient

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator

__all__ = ['AioinjectDependencyProvider']


class AioinjectDependencyProvider(DependencyProvider):
    def __init__(self, container: aioinject.Container | None = None) -> None:
        self._container = container or aioinject.Container()
        self._exit_stack = AsyncExitStack()

    def register(self, *providers: Provider[Any]) -> None:
        self._container.register(*[self._map_provider(provider) for provider in providers])

    @contextmanager
    def override(self, provider: Provider[Any]) -> Iterator[None]:
        override_provider = self._map_provider(provider)
        with self._container.override(override_provider):
            yield

    @asynccontextmanager
    async def _lifespan(self) -> AsyncIterator[None]:
        async with self._container:
            yield

    def _context(self) -> InjectionContext:
        if current_context := aioinject_context.get(None):
            return cast(InjectionContext, nullcontext(current_context))
        return cast(InjectionContext, self._container.context())

    def _map_provider(self, provider: Provider[Any]) -> aioinject.Provider[Any]:  # noqa: PLR6301
        provider_type_map: dict[type[Provider[Any]], type[aioinject.Provider[Any]]] = {
            Transient: aioinject.Transient,
            Scoped: aioinject.Scoped,
            Singleton: aioinject.Singleton,
            Object: aioinject.Object,
        }
        try:
            provider_type = provider_type_map[type(provider)]
        except KeyError:
            msg = 'Unknown provider type'
            raise NotImplementedError(msg) from None

        return provider_type(provider.impl, provider.type_)  # type: ignore[call-arg]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from waku.ext.mediator.handlers.handler import Request, RequestHandlerType


__all__ = [
    'MediatorError',
    'RequestHandlerAlreadyRegistered',
    'RequestHandlerNotFound',
]


class MediatorError(Exception):
    pass


class RequestHandlerAlreadyRegistered(MediatorError, KeyError):  # noqa: N818
    request_type: type[Request[Any]]
    handler_type: RequestHandlerType[Any, Any]

    def __init__(
        self,
        msg: str,
        request_type: type[Request[Any]],
        handler_type: RequestHandlerType[Any, Any],
    ) -> None:
        super().__init__(msg)
        self.request_type = request_type
        self.handler_type = handler_type


class RequestHandlerNotFound(MediatorError, TypeError):  # noqa: N818
    request_type: type[Request[Any]]

    def __init__(self, msg: str, request_type: type[Request[Any]]) -> None:
        super().__init__(msg)
        self.request_type = request_type



from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import Any, ClassVar

from waku.di import DependencyProvider, InjectionContext, Provider


class DummyDI(DependencyProvider):
    _providers: ClassVar = {}

    def register(self, *providers: Provider[Any]) -> None:
        for provider in providers:
            self._providers[provider.type_] = provider

    def override(self, provider: Provider[Any]) -> AbstractContextManager[None]:
        raise NotImplementedError

    def _lifespan(self) -> AbstractAsyncContextManager[None]:
        raise NotImplementedError

    def _context(self) -> InjectionContext:
        raise NotImplementedError



import pytest

from waku.ext.validation import ValidationRule
from waku.ext.validation.rules import DependenciesAccessible, DIScopeMismatch


@pytest.fixture
def rules() -> list[ValidationRule]:
    return [DependenciesAccessible(), DIScopeMismatch()]



from waku.ext.mediator.extensions import MediatorAppExtension, MediatorModuleExtension
from waku.ext.mediator.handlers.handler import Request, RequestHandler, Response
from waku.ext.mediator.handlers.map import RequestMap
from waku.ext.mediator.mediator import Mediator
from waku.ext.mediator.middlewares import MiddlewareChain

__all__ = [
    'Mediator',
    'MediatorAppExtension',
    'MediatorModuleExtension',
    'MiddlewareChain',
    'Request',
    'RequestHandler',
    'RequestMap',
    'Response',
]



from __future__ import annotations

import abc
import uuid
from dataclasses import dataclass, field
from typing import Any, Generic, TypeAlias, TypeVar

__all__ = [
    'Request',
    'RequestHandler',
    'RequestHandlerType',
    'RequestT',
    'Response',
    'ResponseT',
]

RequestT = TypeVar('RequestT', bound='Request[Any]', contravariant=True)  # noqa: PLC0105
ResponseT = TypeVar('ResponseT', bound='Response | None', covariant=True)  # noqa: PLC0105


@dataclass(frozen=True, kw_only=True)
class Request(Generic[ResponseT]):
    """Base class for request-type objects."""

    request_id: uuid.UUID = field(default_factory=uuid.uuid4)


@dataclass(frozen=True, kw_only=True)
class Response:
    """Base class for response type objects."""


class RequestHandler(abc.ABC, Generic[RequestT, ResponseT]):
    @abc.abstractmethod
    async def handle(self, request: RequestT) -> ResponseT:
        raise NotImplementedError


RequestHandlerType: TypeAlias = type[RequestHandler[RequestT, ResponseT]]



from __future__ import annotations

import functools
import inspect
import typing
from typing import TYPE_CHECKING, Any

from waku.di import Object, Provider, Scoped, Singleton, Transient
from waku.ext.validation._abc import ValidationRule
from waku.ext.validation._errors import ValidationError

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from types import UnionType

    from waku import Application, Module
    from waku.ext.validation._extension import ValidationContext

__all__ = [
    'DIScopeMismatch',
    'DependenciesAccessible',
]


class DependenciesAccessible(ValidationRule):
    """Check if all dependencies of providers are accessible."""

    def validate(self, context: ValidationContext) -> ValidationError | None:  # noqa: PLR6301
        # fmt: off
        global_providers = {
            provider
            for module in context.app.iter_submodules()
            for provider in _module_provided_types(module)
            if module.is_global
        }
        # fmt: on

        for module in context.app.iter_submodules():
            for provider in module.providers:
                for dependency in _provider_dependencies(provider=provider):
                    # 1. Dep available globally or provided by current module
                    if dependency in global_providers or dependency in _module_provided_types(module):
                        continue
                    # 2. Dep provided by any of imported modules
                    dependency_accessible = any(
                        _is_exported_dependency(dependency, imported_module)
                        for imported_module in module.iter_submodules()
                    )
                    if not dependency_accessible:
                        err_msg = f"{module!r} depends on {dependency!r} but it's not accessible to it"
                        return ValidationError(err_msg)

        return None


class DIScopeMismatch(ValidationRule):
    """Check if Singleton and Object providers don't depend on Scoped and Transient ones."""

    def validate(self, context: ValidationContext) -> ValidationError | None:
        lifespan_scoped: UnionType = Singleton | Object
        scope_scoped: UnionType = Scoped | Transient

        providers = {provider.type_: provider for provider in self._all_providers(context.app)}
        for provider in self._all_providers(context.app):
            for dependency_type in _provider_dependencies(provider=provider):
                dependency_provider = providers[dependency_type]
                if isinstance(provider, Object):
                    continue

                if (
                    isinstance(provider, lifespan_scoped)
                    and isinstance(dependency_provider, scope_scoped)
                    and not isinstance(dependency_provider, lifespan_scoped)
                ):
                    err_msg = (
                        f'{provider!r} depends on {dependency_provider!r}\n'
                        f'Consider either:\n'
                        f'  - Making {provider!r} {Scoped.__name__} or {Transient.__name__}\n'
                        f'  - Making {dependency_provider!r} {Singleton.__name__} or {Object.__name__}'
                    )
                    return ValidationError(err_msg)

        return None

    def _all_providers(self, app: Application) -> Iterator[Provider[Any]]:  # noqa: PLR6301
        for module in app.iter_submodules():
            yield from module.providers


@functools.cache
def _provider_dependencies(provider: Provider[Any]) -> Sequence[type[object]]:
    if isinstance(provider, Object):
        return ()

    impl = provider.impl
    if inspect.isclass(impl):
        impl = impl.__init__

    params = typing.get_type_hints(impl)
    params.pop('return', None)
    return tuple(params.values())


def _is_exported_dependency(dependency: type[object], module: Module) -> bool:
    # fmt: off
    return (
        dependency in _module_provided_types(module)
        and dependency in module.exports
    )
    # fmt: on


@functools.cache
def _module_provided_types(module: Module | Application) -> set[type[object]]:
    return {provider.type_ for provider in module.providers}



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final

from waku.extensions import ModuleExtension, OnModuleInit

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from waku.di import Provider

__all__ = ['Module']


class Module:
    def __init__(
        self,
        name: str,
        *,
        providers: Sequence[Provider[Any]] = (),
        imports: Sequence[Module] = (),
        exports: Sequence[type[object] | Module] = (),
        extensions: Sequence[ModuleExtension] = (),
        is_global: bool = False,
    ) -> None:
        self.name: Final = name
        self.providers: Final = providers
        self.imports: Final = imports
        self.exports: Final = exports
        self.module_extensions: Final = extensions
        self.is_global: Final = is_global

        self._init_extensions()

    def iter_submodules(self) -> Iterable[Module]:
        yield self
        yield from self.imports

    def _init_extensions(self) -> None:
        for ext in self.module_extensions:
            if isinstance(ext, OnModuleInit):
                ext.on_module_init(self)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'Module[{self.name}]'



from __future__ import annotations

from collections.abc import Callable
from contextlib import AbstractAsyncContextManager, AsyncExitStack, asynccontextmanager, nullcontext
from typing import TYPE_CHECKING, Any, Final, Self, TypeAlias

from waku.di import DependencyProvider, Object
from waku.ext import DEFAULT_EXTENSIONS
from waku.extensions import (
    ApplicationExtension,
    OnApplicationInit,
    OnApplicationShutdown,
    OnApplicationStartup,
)
from waku.module import Module

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Sequence
    from types import TracebackType

    from waku.di import Provider

__all__ = [
    'Application',
    'ApplicationLifespan',
]

ApplicationLifespan: TypeAlias = (
    Callable[['Application'], AbstractAsyncContextManager[None]] | AbstractAsyncContextManager[None]
)


class Application(Module):
    def __init__(
        self,
        name: str,
        *,
        modules: Sequence[Module],
        dependency_provider: DependencyProvider,
        providers: Sequence[Provider[Any]] = (),
        extensions: Sequence[ApplicationExtension] = DEFAULT_EXTENSIONS,
        lifespan: Sequence[ApplicationLifespan] = (),
    ) -> None:
        super().__init__(
            name=name,
            providers=providers,
            imports=modules,
            is_global=True,
        )

        self.extensions: Final = extensions
        self.dependency_provider: Final = dependency_provider

        self._lifespan_managers: list[ApplicationLifespan] = list(lifespan) or [nullcontext()]
        self._exit_stack = AsyncExitStack()

        self._init()

    @property
    def modules(self) -> Sequence[Module]:
        return self.imports

    def _init(self) -> None:
        self.dependency_provider.register(Object(self))
        self.dependency_provider.register(Object(self.dependency_provider, DependencyProvider))

        for module in self.iter_submodules():
            self.dependency_provider.register(*module.providers)

        for ext in self.extensions:
            if isinstance(ext, OnApplicationInit):
                ext.on_app_init(self)

    @asynccontextmanager
    async def _lifespan(self) -> AsyncIterator[None]:
        on_startup = [ext for ext in self.extensions if isinstance(ext, OnApplicationStartup)]
        on_shutdown = [ext for ext in self.extensions if isinstance(ext, OnApplicationShutdown)]

        for ext in on_shutdown[::-1]:
            self._exit_stack.push_async_callback(ext.on_app_shutdown, self)

        for manager in self._lifespan_managers:
            if not isinstance(manager, AbstractAsyncContextManager):
                manager = manager(self)  # noqa: PLW2901
            await self._exit_stack.enter_async_context(manager)

            for on_startup_ext in on_startup:
                await on_startup_ext.on_app_startup(self)

            yield

    async def __aenter__(self) -> Self:
        await self._exit_stack.__aenter__()
        await self._exit_stack.enter_async_context(self._lifespan())
        await self._exit_stack.enter_async_context(self.dependency_provider)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    def __repr__(self) -> str:
        return f'Application[{self.name}]'



from __future__ import annotations

import collections.abc
import inspect
import sys
import typing
from contextlib import contextmanager
from dataclasses import dataclass
from typing import NewType

from waku.di._markers import Inject

if typing.TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from waku.di._types import FactoryType

__all__ = [
    'Dependency',
    'clear_wrapper',
    'collect_dependencies',
    'guess_return_type',
]

_T = typing.TypeVar('_T')
_F = typing.Callable[..., typing.Any]

_GENERATORS: typing.Final = {
    collections.abc.Generator,
    collections.abc.Iterator,
}
_ASYNC_GENERATORS: typing.Final = {
    collections.abc.AsyncGenerator,
    collections.abc.AsyncIterator,
}

_sentinel = object()


def guess_return_type(factory: FactoryType[_T]) -> type[_T]:
    unwrapped = inspect.unwrap(factory)

    origin = typing.get_origin(factory)
    is_generic = origin and inspect.isclass(origin)
    if inspect.isclass(factory) or is_generic:
        return typing.cast(type[_T], factory)

    try:
        return_type = _get_type_hints(unwrapped)['return']
    except KeyError as e:
        msg = f'Factory {factory.__qualname__} does not specify return type.'
        raise ValueError(msg) from e
    except TypeError:
        if isinstance(factory, NewType):
            msg = f'Implementation should be added to provider for type <{factory}> created via NewType.'
            raise ValueError(msg) from None  # noqa: TRY004
        raise
    except NameError:
        # handle future annotations.
        # functions might have dependencies in them
        # and we don't have the container context here so
        # we can't call _get_type_hints
        ret_annotation = unwrapped.__annotations__['return']

        try:
            return_type = _get_return_annotation(ret_annotation, _get_fn_ns(unwrapped))
        except NameError as e:
            msg = f"Factory {factory.__qualname__} does not specify return type. Or it's type is not defined yet."
            raise ValueError(msg) from e

    if origin := typing.get_origin(return_type):
        args = typing.get_args(return_type)

        is_async_gen = origin in _ASYNC_GENERATORS and inspect.isasyncgenfunction(unwrapped)
        is_sync_gen = origin in _GENERATORS and inspect.isgeneratorfunction(
            unwrapped,
        )
        if is_async_gen or is_sync_gen:
            return_type = args[0]

    # classmethod returning `typing.Self`
    if return_type == typing.Self and (self_cls := getattr(factory, '__self__', None)):
        return typing.cast(type[_T], self_cls)

    return typing.cast(type[_T], return_type)


def _get_type_hints(
    obj: typing.Any,
    context: dict[str, type[typing.Any]] | None = None,
) -> dict[str, typing.Any]:
    if not context:
        context = {}
    return typing.get_type_hints(obj, include_extras=True, localns=context)


def _get_fn_ns(fn: collections.abc.Callable[..., typing.Any]) -> dict[str, typing.Any]:
    return getattr(sys.modules.get(fn.__module__, None), '__dict__', {})


def _get_return_annotation(
    ret_annotation: str,
    context: dict[str, typing.Any],
) -> type[typing.Any]:
    return eval(ret_annotation, context)  # type:ignore[no-any-return] # noqa: S307


@dataclass(slots=True, kw_only=True, frozen=True)
class Dependency(typing.Generic[_T]):
    name: str
    type_: type[_T]


def collect_dependencies(
    dependent: typing.Callable[..., object] | dict[str, typing.Any],
    ctx: dict[str, type[typing.Any]] | None = None,
) -> typing.Iterable[Dependency[object]]:
    if not isinstance(dependent, dict):
        with _remove_annotation(dependent.__annotations__, 'return'):
            type_hints = _get_type_hints(dependent, context=ctx)
    else:
        type_hints = dependent
    for name, hint in type_hints.items():
        dep_type, args = _get_annotation_args(hint)
        inject_marker = _find_inject_marker_in_annotation_args(args)
        if inject_marker is None:
            continue

        yield Dependency(name=name, type_=dep_type)


def _get_annotation_args(type_hint: typing.Any) -> tuple[type, tuple[typing.Any, ...]]:
    try:
        dep_type, *args = typing.get_args(type_hint)
    except ValueError:
        dep_type, args = type_hint, []
    return dep_type, tuple(args)


def _find_inject_marker_in_annotation_args(args: tuple[typing.Any, ...]) -> Inject | None:
    for arg in args:
        try:
            if issubclass(arg, Inject):
                return Inject()
        except TypeError:
            pass

        if isinstance(arg, Inject):
            return arg
    return None


def clear_wrapper(wrapper: _F) -> _F:
    inject_annotations = _get_inject_annotations(wrapper)
    signature = inspect.signature(wrapper)
    new_params = tuple(p for p in signature.parameters.values() if p.name not in inject_annotations)
    wrapper.__signature__ = signature.replace(  # type: ignore[attr-defined]
        parameters=new_params,
    )
    for name in inject_annotations:
        del wrapper.__annotations__[name]
    return wrapper


def _get_inject_annotations(function: Callable[..., typing.Any]) -> dict[str, typing.Any]:
    with _remove_annotation(function.__annotations__, 'return'):
        return {
            name: annotation
            for name, annotation in typing.get_type_hints(
                function,
                include_extras=True,
            ).items()
            if any(isinstance(arg, Inject) or arg is Inject for arg in typing.get_args(annotation))
        }


@contextmanager
def _remove_annotation(annotations: dict[str, typing.Any], name: str) -> Iterator[None]:
    annotation = annotations.pop(name, _sentinel)
    yield
    if annotation is not _sentinel:
        annotations[name] = annotation



from waku.di._inject import inject
from waku.di._markers import Injected
from waku.di._providers import DependencyProvider, InjectionContext, Object, Provider, Scoped, Singleton, Transient

__all__ = [
    'DependencyProvider',
    'Injected',
    'InjectionContext',
    'Object',
    'Provider',
    'Scoped',
    'Singleton',
    'Transient',
    'inject',
]



from __future__ import annotations

from typing import Any, Self, TypeAlias

from waku.ext.mediator.exceptions import RequestHandlerAlreadyRegistered, RequestHandlerNotFound
from waku.ext.mediator.handlers.handler import RequestHandlerType, RequestT, ResponseT

__all__ = [
    'RequestMap',
    'RequestMapRegistry',
]

RequestMapRegistry: TypeAlias = dict[type[RequestT], RequestHandlerType[RequestT, ResponseT]]


class RequestMap:
    def __init__(self) -> None:
        self._registry: RequestMapRegistry[Any, Any] = {}

    def bind(self, request_type: type[RequestT], handler_type: RequestHandlerType[RequestT, ResponseT]) -> Self:
        if request_type in self._registry:
            msg = f'{request_type.__name__} already exists in registry'
            raise RequestHandlerAlreadyRegistered(msg, request_type, handler_type)
        self._registry[request_type] = handler_type
        return self

    def merge(self, other: RequestMap) -> Self:
        for request_type, handler_type in other.registry.items():
            self.bind(request_type, handler_type)
        return self

    @property
    def registry(self) -> RequestMapRegistry[Any, Any]:
        return self._registry

    def __getitem__(self, request_type: type[RequestT]) -> RequestHandlerType[RequestT, ResponseT]:
        try:
            return self._registry[request_type]
        except KeyError as err:
            msg = f'Request handler for {request_type.__name__} request is not registered'
            raise RequestHandlerNotFound(msg, request_type) from err



from __future__ import annotations

from typing import TYPE_CHECKING, Final

from waku.di import DependencyProvider, Object, Scoped
from waku.ext.mediator.handlers.map import RequestMap
from waku.ext.mediator.mediator import Mediator
from waku.extensions import OnApplicationInit, OnModuleInit

if TYPE_CHECKING:
    from waku.application import Application
    from waku.ext.mediator.handlers.dispatcher import RequestDispatcher
    from waku.ext.mediator.middlewares import MiddlewareChain
    from waku.module import Module


__all__ = [
    'MediatorAppExtension',
    'MediatorModuleExtension',
]


class MediatorAppExtension(OnApplicationInit):
    def __init__(
        self,
        middleware_chain: MiddlewareChain | None = None,
        dispatcher_class: type[RequestDispatcher] | None = None,
    ) -> None:
        self._middleware_chain = middleware_chain
        self._dispatcher_class = dispatcher_class

    def on_app_init(self, app: Application) -> None:
        dp = app.dependency_provider

        request_map = RequestMap()
        for module in app.modules:
            module_ext = next(
                (ext for ext in module.module_extensions if isinstance(ext, MediatorModuleExtension)), None
            )
            if not module_ext:
                continue

            request_map.merge(module_ext.request_map)

        for handler_type in request_map.registry.values():
            dp.register(Scoped(handler_type))

        mediator = self._build_mediator(dp, request_map)
        dp.register(Object(mediator, Mediator))

    def _build_mediator(self, dependency_provider: DependencyProvider, request_map: RequestMap) -> Mediator:
        return Mediator(
            request_map=request_map,
            dependency_provider=dependency_provider,
            middleware_chain=self._middleware_chain,
            dispatcher_class=self._dispatcher_class,
        )


class MediatorModuleExtension(OnModuleInit):
    def __init__(self, request_map: RequestMap, /) -> None:
        self.request_map: Final = request_map

    def on_module_init(self, module: Module) -> None:
        pass



from __future__ import annotations

import functools
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, Generic, Self, TypeAlias

from waku.ext.mediator.handlers.handler import RequestT, ResponseT

if TYPE_CHECKING:
    from collections.abc import Sequence


__all__ = [
    'AnyMiddleware',
    'HandleType',
    'Middleware',
    'MiddlewareChain',
]

HandleType: TypeAlias = Callable[[RequestT], Awaitable[ResponseT]]


class Middleware(ABC, Generic[RequestT, ResponseT]):
    @abstractmethod
    async def __call__(
        self,
        request: RequestT,
        handle: HandleType[RequestT, ResponseT],
    ) -> ResponseT: ...


AnyMiddleware = Middleware[Any, Any]


class MiddlewareChain:
    def __init__(self, middlewares: Sequence[AnyMiddleware] = ()) -> None:
        self._middlewares: list[AnyMiddleware] = list(middlewares)

    def add(self, middleware: AnyMiddleware) -> Self:
        self._middlewares.append(middleware)
        return self

    def wrap(self, handle: HandleType[RequestT, ResponseT]) -> HandleType[RequestT, ResponseT]:
        original_handle = handle
        for middleware in reversed(self._middlewares):
            handle = functools.update_wrapper(
                wrapper=functools.partial(middleware.__call__, handle=handle),
                wrapped=original_handle,
            )
        return handle



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, TypeAlias, runtime_checkable

if TYPE_CHECKING:
    from waku.application import Application
    from waku.module import Module


__all__ = [
    'ApplicationExtension',
    'ModuleExtension',
    'OnApplicationInit',
    'OnApplicationShutdown',
    'OnApplicationStartup',
    'OnModuleInit',
]


@runtime_checkable
class OnApplicationInit(Protocol):
    __slots__ = ()

    def on_app_init(self, app: Application) -> None: ...


@runtime_checkable
class OnApplicationStartup(Protocol):
    __slots__ = ()

    async def on_app_startup(self, app: Application) -> None: ...


@runtime_checkable
class OnApplicationShutdown(Protocol):
    __slots__ = ()

    async def on_app_shutdown(self, app: Application) -> None: ...


ApplicationExtension: TypeAlias = OnApplicationInit | OnApplicationStartup | OnApplicationShutdown


@runtime_checkable
class OnModuleInit(Protocol):
    __slots__ = ()

    def on_module_init(self, module: Module) -> None: ...


ModuleExtension: TypeAlias = OnModuleInit



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Final

from litestar.middleware import MiddlewareProtocol
from litestar.plugins import InitPluginProtocol

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from litestar import Litestar
    from litestar.config.app import AppConfig
    from litestar.types import ASGIApp, Receive, Scope, Send

    from waku.application import Application

__all__ = [
    'ApplicationMiddleware',
    'ApplicationPlugin',
]

_STATE_KEY: Final = '__waku_application__'
_SCOPE_CONTEXT_KEY: Final = '__waku_injection_context__'


class ApplicationMiddleware(MiddlewareProtocol):
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        app: Litestar = scope['app']
        application: Application = app.state[_STATE_KEY]

        async with application.dependency_provider.context() as ctx:
            scope[_SCOPE_CONTEXT_KEY] = ctx  # type: ignore[typeddict-unknown-key]
            await self.app(scope, receive, send)


async def _after_exception(exception: BaseException, scope: Scope) -> None:
    if _SCOPE_CONTEXT_KEY in scope:
        await scope[_SCOPE_CONTEXT_KEY].__aexit__(  # type: ignore[typeddict-item]
            type(exception),
            exception,
            exception.__traceback__,
        )


class ApplicationPlugin(InitPluginProtocol):
    def __init__(self, application: Application) -> None:
        self._application = application

    @contextlib.asynccontextmanager
    async def _lifespan(self, _: Litestar) -> AsyncIterator[None]:
        async with self._application:
            yield

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.state[_STATE_KEY] = self._application
        app_config.middleware.append(ApplicationMiddleware)
        app_config.lifespan.append(self._lifespan)
        app_config.after_exception.append(_after_exception)
        return app_config



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic

from waku.ext.mediator.handlers.handler import Request, RequestHandlerType, ResponseT
from waku.ext.mediator.middlewares import MiddlewareChain

if TYPE_CHECKING:
    from waku.di import DependencyProvider
    from waku.ext.mediator.handlers.map import RequestMap


__all__ = [
    'RequestDispatchResult',
    'RequestDispatcher',
]


@dataclass(frozen=True, kw_only=True)
class RequestDispatchResult(Generic[ResponseT]):
    response: ResponseT
    # TODO: events  # noqa: FIX002


class RequestDispatcher:
    def __init__(
        self,
        request_map: RequestMap,
        *,
        dependency_provider: DependencyProvider,
        middleware_chain: MiddlewareChain | None = None,
    ) -> None:
        self._request_map = request_map
        self._dependency_provider = dependency_provider
        self._middleware_chain = middleware_chain or MiddlewareChain()

    async def dispatch(self, request: Request[ResponseT]) -> RequestDispatchResult[ResponseT]:
        handler_type: RequestHandlerType[Any, Any] = self._request_map[type(request)]

        async with self._dependency_provider.context() as ctx:
            handler = await ctx.resolve(handler_type)
            wrapped_handler = self._middleware_chain.wrap(handler.handle)
            response = await wrapped_handler(request)

        return RequestDispatchResult(response=response)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated, Generic, TypeAlias, TypeVar

__all__ = ['Inject', 'Injected']

_T = TypeVar('_T')


@dataclass(slots=True)
class Inject:
    pass


if TYPE_CHECKING:
    Injected: TypeAlias = Annotated[_T, Inject]

else:

    class Injected(Generic[_T]):
        def __class_getitem__(cls, item: object) -> object:
            return Annotated[item, Inject]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, MutableMapping

    from waku.application import Application

    Scope: TypeAlias = MutableMapping[str, Any]
    Message: TypeAlias = MutableMapping[str, Any]
    Receive: TypeAlias = Callable[[], Awaitable[Message]]
    Send: TypeAlias = Callable[[Message], Awaitable[None]]

    ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]


__all__ = ['ApplicationMiddleware']


class ApplicationMiddleware:
    def __init__(self, app: ASGIApp, *, application: Application) -> None:
        self.app = app
        self._application = application

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with self._application.dependency_provider.context():
            await self.app(scope, receive, send)



from typing import Protocol

from waku.ext.validation._errors import ValidationError
from waku.ext.validation._extension import ValidationContext


class ValidationRule(Protocol):
    def validate(self, context: ValidationContext) -> ValidationError | None: ...
